# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
DXFile Handler
**************

This remote file handler is a Python file-like object.
'''

from __future__ import print_function, unicode_literals, division, absolute_import

import os, sys, traceback, copy, time
from io import BytesIO
import math
import mmap
from threading import Lock
from multiprocessing import cpu_count

import dxpy
from . import DXDataObject
from ..exceptions import DXFileError, DXIncompleteReadsError
from ..utils import warn
from ..utils.resolver import object_exists_in_project
from ..compat import md5_hasher, basestring


DXFILE_HTTP_THREADS = min(cpu_count(), 8)
MIN_BUFFER_SIZE = 1024*1024
DEFAULT_BUFFER_SIZE = 1024*1024*16
if dxpy.JOB_ID:
    # Increase HTTP request buffer size when we are running within the
    # platform.
    DEFAULT_BUFFER_SIZE = 1024*1024*96

MD5_READ_CHUNK_SIZE = 1024*1024*4
FILE_REQUEST_TIMEOUT = 60
PART_UPLOAD_REQUEST_TIMEOUT_SECONDS = 300


def _validate_headers(headers):
    for key, value in headers.items():
        if not isinstance(key, basestring):
            raise ValueError("Expected key %r of headers to be a string" % (key,))
        if not isinstance(value, basestring):
            raise ValueError("Expected value %r of headers (associated with key %r) to be a string"
                             % (value, key))
    return headers


def _readable_part_size(num_bytes):
    "Returns the file size in readable form."
    B = num_bytes
    KB = float(1024)
    MB = float(KB * 1024)
    GB = float(MB * 1024)
    TB = float(GB * 1024)

    if B < KB:
        return '{0} {1}'.format(B, 'bytes' if B != 1 else 'byte')
    elif KB <= B < MB:
        return '{0:.2f} KiB'.format(B/KB)
    elif MB <= B < GB:
        return '{0:.2f} MiB'.format(B/MB)
    elif GB <= B < TB:
        return '{0:.2f} GiB'.format(B/GB)
    elif TB <= B:
        return '{0:.2f} TiB'.format(B/TB)


def _get_write_buf_size(buffer_size_hint, file_upload_params, expected_file_size, file_is_mmapd=False):
    max_num_parts = file_upload_params['maximumNumParts']
    min_part_size = file_upload_params['minimumPartSize']
    max_part_size = file_upload_params['maximumPartSize']
    max_file_size = file_upload_params['maximumFileSize']

    if expected_file_size is not None and expected_file_size > max_file_size:
        raise DXFileError("Size of file exceeds maximum of {}".format(_readable_part_size(max_file_size)))

    min_buffer_size = min_part_size
    if expected_file_size is not None:
        # Raise buffer size (for files exceeding DEFAULT_BUFFER_SIZE
        # * the maximium parts allowed bytes) in order to prevent us
        # from exceeding the configured parts limit.
        min_buffer_size = max(min_buffer_size, int(math.ceil(float(expected_file_size) / max_num_parts)))
    max_buffer_size = max_part_size

    assert min_buffer_size <= max_buffer_size

    if file_is_mmapd:
        # If file is mmapd, force the eventual result to be a
        # multiple of the allocation granularity by rounding all of
        # buffer_size, min_buffer_size, and max_buffer_size to a
        # nearby multiple of the allocation granularity (below, the
        # final buffer size will be one of these).
        if min_buffer_size % mmap.ALLOCATIONGRANULARITY != 0:
            min_buffer_size += mmap.ALLOCATIONGRANULARITY - min_buffer_size % mmap.ALLOCATIONGRANULARITY
        if max_buffer_size % mmap.ALLOCATIONGRANULARITY != 0:
            max_buffer_size -= max_buffer_size % mmap.ALLOCATIONGRANULARITY
        buffer_size_hint = buffer_size_hint - buffer_size_hint % mmap.ALLOCATIONGRANULARITY
    else:
        buffer_size_hint = buffer_size_hint

    # Use the user-specified hint if it is a permissible size
    # (satisfies API and large enough to upload file of advertised
    # size). Otherwise, select the closest size that is permissible.
    buffer_size = buffer_size_hint
    buffer_size = max(buffer_size, min_buffer_size)
    buffer_size = min(buffer_size, max_buffer_size)

    if expected_file_size is not None and (buffer_size * max_num_parts < expected_file_size):
        raise AssertionError("part size would be too small to upload the requested number of bytes")

    if file_is_mmapd and buffer_size % mmap.ALLOCATIONGRANULARITY != 0:
        raise AssertionError('part size will not be accepted by mmap')

    return buffer_size


class DXFile(DXDataObject):
    '''Remote file object handler.

    :param dxid: Object ID
    :type dxid: string
    :param project: Project ID
    :type project: string
    :param mode: One of "r", "w", or "a" for read, write, and append modes, respectively.
                 Use "b" for binary mode. For example, "rb" means open a file for reading
                 in binary mode.
    :type mode: string

    .. note:: The attribute values below are current as of the last time
              :meth:`~dxpy.bindings.DXDataObject.describe` was run.
              (Access to any of the below attributes causes
              :meth:`~dxpy.bindings.DXDataObject.describe` to be called
              if it has never been called before.)

    .. py:attribute:: media

       String containing the Internet Media Type (also known as MIME type
       or Content-type) of the file.

    .. automethod:: _new

    '''

    _class = "file"

    _describe = staticmethod(dxpy.api.file_describe)
    _add_types = staticmethod(dxpy.api.file_add_types)
    _remove_types = staticmethod(dxpy.api.file_remove_types)
    _get_details = staticmethod(dxpy.api.file_get_details)
    _set_details = staticmethod(dxpy.api.file_set_details)
    _set_visibility = staticmethod(dxpy.api.file_set_visibility)
    _rename = staticmethod(dxpy.api.file_rename)
    _set_properties = staticmethod(dxpy.api.file_set_properties)
    _add_tags = staticmethod(dxpy.api.file_add_tags)
    _remove_tags = staticmethod(dxpy.api.file_remove_tags)
    _close = staticmethod(dxpy.api.file_close)
    _list_projects = staticmethod(dxpy.api.file_list_projects)

    _http_threadpool_size = DXFILE_HTTP_THREADS
    _http_threadpool = dxpy.utils.get_futures_threadpool(max_workers=_http_threadpool_size)

    NO_PROJECT_HINT = 'NO_PROJECT_HINT'

    @classmethod
    def set_http_threadpool_size(cls, num_threads):
        '''

        .. deprecated:: 0.191.0

        '''
        print('set_http_threadpool_size is deprecated')

    def __init__(self, dxid=None, project=None, mode=None, read_buffer_size=DEFAULT_BUFFER_SIZE,
                 write_buffer_size=DEFAULT_BUFFER_SIZE, expected_file_size=None, file_is_mmapd=False):
        """
        :param dxid: Object ID
        :type dxid: string
        :param project: Project ID
        :type project: string
        :param mode: One of "r", "w", or "a" for read, write, and append
            modes, respectively. Add "b" for binary mode.
        :type mode: string
        :param read_buffer_size: size of read buffer in bytes
        :type read_buffer_size: int
        :param write_buffer_size: hint for size of write buffer in
            bytes. A lower or higher value may be used depending on
            region-specific parameters and on the expected file size.
        :type write_buffer_size: int
        :param expected_file_size: size of data that will be written, if
            known
        :type expected_file_size: int
        :param file_is_mmapd: True if input file is mmap'd (if so, the
            write buffer size will be constrained to be a multiple of
            the allocation granularity)
        :type file_is_mmapd: bool
        """
        DXDataObject.__init__(self, dxid=dxid, project=project)

        # By default, a file is created in text mode. This makes a difference
        # in python 3.
        self._binary_mode = False
        if mode is None:
            self._close_on_exit = True
        else:
            if 'b' in mode:
                self._binary_mode = True
                mode = mode.replace("b", "")
            if mode not in ['r', 'w', 'a']:
                raise ValueError("mode must be one of 'r', 'w', or 'a'. Character 'b' may be used in combination (e.g. 'wb').")
            self._close_on_exit = (mode == 'w')
        self._read_buf = BytesIO()
        self._write_buf = BytesIO()

        self._read_bufsize = read_buffer_size

        # Computed lazily later since this depends on the project, and
        # we want to allow the project to be set as late as possible.
        # Call _ensure_write_bufsize to ensure that this is set before
        # trying to read it.
        self._write_bufsize = None

        self._write_buffer_size_hint = write_buffer_size
        self._expected_file_size = expected_file_size
        self._file_is_mmapd = file_is_mmapd

        # These are cached once for all download threads. This saves calls to the apiserver.
        self._download_url, self._download_url_headers, self._download_url_expires = None, None, None

        # This lock protects accesses to the above three variables, ensuring that they would
        # be checked and changed atomically. This protects against thread race conditions.
        self._url_download_mutex = Lock()

        self._request_iterator, self._response_iterator = None, None
        self._http_threadpool_futures = set()

        # Initialize state
        self._pos = 0
        self._file_length = None
        self._cur_part = 1
        self._num_uploaded_parts = 0
        # Cache for object existence in project
        self._exists_in_proj = None  

    def _new(self, dx_hash, media_type=None, **kwargs):
        """
        :param dx_hash: Standard hash populated in :func:`dxpy.bindings.DXDataObject.new()` containing attributes common to all data object classes.
        :type dx_hash: dict
        :param media_type: Internet Media Type
        :type media_type: string

        Creates a new remote file with media type *media_type*, if given.

        """

        if media_type is not None:
            dx_hash["media"] = media_type

        resp = dxpy.api.file_new(dx_hash, **kwargs)
        self.set_ids(resp["id"], dx_hash["project"])

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.flush()
        if self._close_on_exit and self._get_state() == "open":
            self.close()

    def __del__(self):
        '''
        Exceptions raised here in the destructor are IGNORED by Python! We will try and flush data
        here just as a safety measure, but you should not rely on this to flush your data! We will
        be really grumpy and complain if we detect unflushed data here.

        Use a context manager or flush the object explicitly to avoid this.

        In addition, when this is triggered by interpreter shutdown, the thread pool is not
        available, and we will wait for the request queue forever. In this case, we must revert to
        synchronous, in-thread flushing. We don't know how to detect this condition, so we'll use
        that for all destructor events.

        Neither this nor context managers are compatible with kwargs pass-through (so e.g. no
        custom auth).
        '''
        if not hasattr(self, '_write_buf'):
            # This occurs when there is an exception initializing the
            # DXFile object
            return

        if self._write_buf.tell() > 0 or len(self._http_threadpool_futures) > 0:
            warn("=== WARNING! ===")
            warn("There is still unflushed data in the destructor of a DXFile object!")
            warn("We will attempt to flush it now, but if an error were to occur, we could not report it back to you.")
            warn("Your program could fail to flush the data but appear to succeed.")
            warn("Instead, please call flush() or close(), or use the context managed version (e.g., with open_dxfile(ID, mode='w') as f:)")
        try:
            self.flush(multithread=False)
        except Exception as e:
            warn("=== Exception occurred while flushing accumulated file data for %r" % (self._dxid,))
            traceback.print_exception(*sys.exc_info())
            raise

    def __iter__(self):
        _buffer = ""
        if self._binary_mode:
            raise DXFileError("Cannot read lines when file opened in binary mode")
        while True:
            more = self.read(self._read_bufsize)
            if not more:
                break
            _buffer += more
            parts = _buffer.split("\n")
            for i in range(len(parts) - 1):
                yield parts[i]
            # The final piece may be incomplete, so keep it in the buffer
            _buffer = parts[-1]
        if _buffer:
            yield _buffer

    next = next
    __next__ = next

    def set_ids(self, dxid, project=None):
        '''
        :param dxid: Object ID
        :type dxid: string
        :param project: Project ID
        :type project: string

        Discards the currently stored ID and associates the handler with
        *dxid*. As a side effect, it also flushes the buffer for the
        previous file object if the buffer is nonempty.
        '''
        if self._dxid is not None:
            self.flush()

        DXDataObject.set_ids(self, dxid, project)

        # Reset state
        self._pos = 0
        self._file_length = None
        self._cur_part = 1
        self._num_uploaded_parts = 0

    def seek(self, offset, from_what=os.SEEK_SET):
        '''
        :param offset: Position in the file to seek to
        :type offset: integer

        Seeks to *offset* bytes from the beginning of the file.  This is a no-op if the file is open for writing.

        The position is computed from adding *offset* to a reference point; the reference point is selected by the
        *from_what* argument. A *from_what* value of 0 measures from the beginning of the file, 1 uses the current file
        position, and 2 uses the end of the file as the reference point. *from_what* can be omitted and defaults to 0,
        using the beginning of the file as the reference point.
        '''
        if from_what == os.SEEK_SET:
            reference_pos = 0
        elif from_what == os.SEEK_CUR:
            reference_pos = self._pos
        elif from_what == os.SEEK_END:
            if self._file_length == None:
                desc = self.describe()
                self._file_length = int(desc["size"])
            reference_pos = self._file_length
        else:
            raise DXFileError("Invalid value supplied for from_what")

        orig_pos = self._pos
        self._pos = reference_pos + offset

        in_buf = False
        orig_buf_pos = self._read_buf.tell()
        if offset < orig_pos:
            if orig_buf_pos > orig_pos - offset:
                # offset is less than original position but within the buffer
                in_buf = True
        else:
            buf_len = dxpy.utils.string_buffer_length(self._read_buf)
            if buf_len - orig_buf_pos > offset - orig_pos:
                # offset is greater than original position but within the buffer
                in_buf = True

        if in_buf:
            # offset is within the buffer (at least one byte following
            # the offset can be read directly out of the buffer)
            self._read_buf.seek(orig_buf_pos - orig_pos + offset)
        elif offset == orig_pos:
            # This seek is a no-op (the cursor is just past the end of
            # the read buffer and coincides with the desired seek
            # position). We don't have the data ready, but the request
            # for the data starting here is already in flight.
            #
            # Detecting this case helps to optimize for sequential read
            # access patterns.
            pass
        else:
            # offset is outside the buffer-- reset buffer and queues.
            # This is the failsafe behavior
            self._read_buf = BytesIO()
            # TODO: if the offset is within the next response(s), don't throw out the queues
            self._request_iterator, self._response_iterator = None, None

    def tell(self):
        '''
        Returns the current position of the file read cursor.

        Warning: Because of buffering semantics, this value will **not** be accurate when using the line iterator form
        (`for line in file`).
        '''
        return self._pos

    def flush(self, multithread=True, **kwargs):
        '''
        Flushes the internal write buffer.
        '''
        if self._write_buf.tell() > 0:
            data = self._write_buf.getvalue()
            self._write_buf = BytesIO()

            if multithread:
                self._async_upload_part_request(data, index=self._cur_part, **kwargs)
            else:
                self.upload_part(data, self._cur_part, **kwargs)

            self._cur_part += 1

        if len(self._http_threadpool_futures) > 0:
            dxpy.utils.wait_for_all_futures(self._http_threadpool_futures)
            try:
                for future in self._http_threadpool_futures:
                    if future.exception() != None:
                        raise future.exception()
            finally:
                self._http_threadpool_futures = set()

    def _async_upload_part_request(self, *args, **kwargs):
        while len(self._http_threadpool_futures) >= self._http_threadpool_size:
            future = dxpy.utils.wait_for_a_future(self._http_threadpool_futures)
            if future.exception() != None:
                raise future.exception()
            self._http_threadpool_futures.remove(future)

        future = self._http_threadpool.submit(self.upload_part, *args, **kwargs)
        self._http_threadpool_futures.add(future)

    def _ensure_write_bufsize(self, **kwargs):
        if self._write_bufsize is not None:
            return
        file_upload_params = dxpy.api.project_describe(
            self.get_proj_id(),
            {'fields': {'fileUploadParameters': True}},
            **kwargs
        )['fileUploadParameters']
        self._empty_last_part_allowed = file_upload_params['emptyLastPartAllowed']
        self._write_bufsize = _get_write_buf_size(self._write_buffer_size_hint,
                                                  file_upload_params,
                                                  self._expected_file_size,
                                                  self._file_is_mmapd)

    def _write2(self, data, multithread=True, **kwargs):
        '''
        :param data: Data to be written
        :type data: str or mmap object
        :param multithread: If True, sends multiple write requests asynchronously
        :type multithread: boolean

        Writes the data *data* to the file.

        .. note::

            Writing to remote files is append-only. Using :meth:`seek`
            does not affect where the next :meth:`write` will occur.

        '''
        assert(isinstance(data, bytes))

        self._ensure_write_bufsize(**kwargs)

        def write_request(data_for_write_req):
            if multithread:
                self._async_upload_part_request(data_for_write_req, index=self._cur_part, **kwargs)
            else:
                self.upload_part(data_for_write_req, self._cur_part, **kwargs)
            self._cur_part += 1

        if self._write_buf.tell() == 0 and self._write_bufsize == len(data):
            # In the special case of a write that is the same size as
            # our write buffer size, and no unflushed data in the
            # buffer, just directly dispatch the write and bypass the
            # write buffer.
            #
            # This saves a buffer copy, which is especially helpful if
            # 'data' is actually mmap'd from a file.
            #
            # TODO: an additional optimization could be made to allow
            # the last request from an mmap'd upload to take this path
            # too (in general it won't because it's not of length
            # _write_bufsize). This is probably inconsequential though.
            write_request(data)
            return

        remaining_space = self._write_bufsize - self._write_buf.tell()

        if len(data) <= remaining_space:
            self._write_buf.write(data)
        else:
            self._write_buf.write(data[:remaining_space])

            temp_data = self._write_buf.getvalue()
            self._write_buf = BytesIO()
            write_request(temp_data)

            # TODO: check if repeat string splitting is bad for
            # performance when len(data) >> _write_bufsize
            self.write(data[remaining_space:], **kwargs)

    def write(self, data, multithread=True, **kwargs):
        '''
        :param data: Data to be written
        :type data: str or mmap object
        :param multithread: If True, sends multiple write requests asynchronously
        :type multithread: boolean

        Writes the data *data* to the file.

        .. note::

            Writing to remote files is append-only. Using :meth:`seek`
            does not affect where the next :meth:`write` will occur.

        '''
        # In python3, the underlying system methods use the 'bytes' type, not 'string'
        #
        # This is, hopefully, a temporary hack. It is not a good idea for two reasons:
        # 1) Performance, we need to make a pass on the data, and need to allocate
        #    another buffer of similar size
        # 2) The types are wrong. The "bytes" type should be visible to the caller
        #    of the write method, instead of being hidden.

        # Should we throw an exception if the file is opened in binary mode,
        # and the data is unicode/text?
        if isinstance(data, str):
            bt = data.encode("utf-8")
        elif isinstance(data, bytearray):
            bt = bytes(data)
        elif isinstance(data, bytes):
            bt = data
        elif isinstance(data, mmap.mmap):
            bt = bytes(data)
        else:
            raise DXFileError("Invalid type {} for write data argument".format(type(data)))
        assert(isinstance(bt, bytes))
        self._write2(bt, multithread=multithread, **kwargs)

    def closed(self, **kwargs):
        '''
        :returns: Whether the remote file is closed
        :rtype: boolean

        Returns :const:`True` if the remote file is closed and
        :const:`False` otherwise. Note that if it is not closed, it can
        be in either the "open" or "closing" states.
        '''

        return self.describe(fields={'state'}, **kwargs)["state"] == "closed"

    def close(self, block=False, **kwargs):
        '''
        :param block: If True, this function blocks until the remote file has closed.
        :type block: boolean

        Attempts to close the file.

        .. note:: The remote file cannot be closed until all parts have
           been fully uploaded. An exception will be thrown if this is
           not the case.
        '''
        self.flush(**kwargs)

        # Also populates emptyLastPartAllowed
        self._ensure_write_bufsize(**kwargs)

        if self._num_uploaded_parts == 0 and self._empty_last_part_allowed:
            # We haven't uploaded any parts in this session.
            # In case no parts have been uploaded at all and region
            # settings allow last empty part upload, try to upload
            # an empty part (otherwise files with 0 parts cannot be closed).
            try:
                self.upload_part(b'', 1, **kwargs)
            except dxpy.exceptions.InvalidState:
                pass

        if 'report_progress_fn' in kwargs:
            del kwargs['report_progress_fn']

        dxpy.api.file_close(self._dxid, **kwargs)

        if block:
            self._wait_on_close(**kwargs)

    def wait_on_close(self, timeout=3600*24*7, **kwargs):
        '''
        :param timeout: Maximum amount of time to wait (in seconds) until the file is closed.
        :type timeout: integer
        :raises: :exc:`dxpy.exceptions.DXFileError` if the timeout is reached before the remote file has been closed

        Waits until the remote file is closed.
        '''
        self._wait_on_close(timeout, **kwargs)

    def upload_part(self, data, index=None, display_progress=False, report_progress_fn=None, **kwargs):
        """
        :param data: Data to be uploaded in this part
        :type data: str or mmap object, bytes on python3
        :param index: Index of part to be uploaded; must be in [1, 10000]
        :type index: integer
        :param display_progress: Whether to print "." to stderr when done
        :type display_progress: boolean
        :param report_progress_fn: Optional: a function to call that takes in two arguments (self, # bytes transmitted)
        :type report_progress_fn: function or None
        :raises: :exc:`dxpy.exceptions.DXFileError` if *index* is given and is not in the correct range, :exc:`urllib3.exceptions.HTTPError` if upload fails

        Uploads the data in *data* as part number *index* for the
        associated file. If no value for *index* is given, *index*
        defaults to 1. This probably only makes sense if this is the
        only part to be uploaded.
        """
        assert(isinstance(data, bytes))

        req_input = {}
        if index is not None:
            req_input["index"] = int(index)

        md5 = md5_hasher()
        if hasattr(data, 'seek') and hasattr(data, 'tell'):
            # data is a buffer; record initial position (so we can rewind back)
            rewind_input_buffer_offset = data.tell()
            while True:
                bytes_read = data.read(MD5_READ_CHUNK_SIZE)
                if bytes_read:
                    md5.update(bytes_read)
                else:
                    break
            # rewind the buffer to original position
            data.seek(rewind_input_buffer_offset)
        else:
            md5.update(data)

        req_input["md5"] = md5.hexdigest()
        req_input["size"] = len(data)

        def get_upload_url_and_headers():
            # This function is called from within a retry loop, so to avoid amplifying the number of retries
            # geometrically, we decrease the allowed number of retries for the nested API call every time.
            if 'max_retries' not in kwargs:
                kwargs['max_retries'] = dxpy.DEFAULT_RETRIES
            elif kwargs['max_retries'] > 0:
                kwargs['max_retries'] -= 1

            if "timeout" not in kwargs:
                kwargs["timeout"] = FILE_REQUEST_TIMEOUT

            resp = dxpy.api.file_upload(self._dxid, req_input, **kwargs)
            url = resp["url"]
            return url, _validate_headers(resp.get("headers", {}))

        # The file upload API requires us to get a pre-authenticated upload URL (and headers for it) every time we
        # attempt an upload. Because DXHTTPRequest will retry requests under retryable conditions, we give it a callback
        # to ask us for a new upload URL every time it attempts a request (instead of giving them directly).

        dxpy.DXHTTPRequest(get_upload_url_and_headers,
                           data,
                           jsonify_data=False,
                           prepend_srv=False,
                           always_retry=True,
                           timeout=PART_UPLOAD_REQUEST_TIMEOUT_SECONDS,
                           auth=None,
                           method='PUT')

        self._num_uploaded_parts += 1

        if display_progress:
            warn(".")

        if report_progress_fn is not None:
            report_progress_fn(self, len(data))

    def wait_until_parts_uploaded(self, **kwargs):
        self._wait_until_parts_uploaded(**kwargs)


    def get_download_url(self, duration=None, preauthenticated=False, filename=None, project=None, **kwargs):
        """
        :param duration: number of seconds for which the generated URL will be
            valid, should only be specified when preauthenticated is True
        :type duration: int
        :param preauthenticated: if True, generates a 'preauthenticated'
            download URL, which embeds authentication info in the URL and does
            not require additional headers
        :type preauthenticated: bool
        :param filename: desired filename of the downloaded file
        :type filename: str
        :param project: ID of a project containing the file (the download URL
            will be associated with this project, and this may affect which
            billing account is billed for this download).
            If no project is specified, an attempt will be made to verify if the file is
            in the project from the DXFile handler (as specified by the user or
            the current project stored in dxpy.WORKSPACE_ID). Otherwise, no hint is supplied.
            This fall back behavior does not happen inside a job environment.
            A non preauthenticated URL is only valid as long as the user has
            access to that project and the project contains that file.
        :type project: str
        :returns: download URL and dict containing HTTP headers to be supplied
            with the request
        :rtype: tuple (str, dict)
        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if a project context was
            given and the file was not found in that project context.
        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if no project context was
            given and the file was not found in any projects.

        Obtains a URL that can be used to directly download the associated
        file.

        """
        with self._url_download_mutex:
            # Only generate URL if not already cached or expired
            if self._download_url is None or self._download_url_expires < time.time():
                args = {"preauthenticated": preauthenticated}
                if duration is not None:
                    args["duration"] = duration
                if filename is not None:
                    args["filename"] = filename

                # If project=None, we fall back to the project attached to this handler
                # (if any). If this is supplied, it's treated as a hint: if it's a
                # project in which this file exists, it's passed on to the
                # apiserver. Otherwise, NO hint is supplied. In principle supplying a
                # project in the handler that doesn't contain this file ought to be an
                # error, but it's this way for backwards compatibility. We don't know
                # who might be doing downloads and creating handlers without being
                # careful that the project encoded in the handler contains the file
                # being downloaded. They may now rely on such behavior.
                if project is None and 'DX_JOB_ID' not in os.environ:
                    project_from_handler = self.get_proj_id()
                    # object_exists_in_project will call /file-xxxx/describe, which is skipped if the URL is cached
                    if project_from_handler and object_exists_in_project(self.get_id(), project_from_handler):
                        project = project_from_handler

                if project is not None and project is not DXFile.NO_PROJECT_HINT:
                    args["project"] = project

                # Test hook to write 'project' argument passed to API call to a
                # local file
                if '_DX_DUMP_BILLED_PROJECT' in os.environ:
                    with open(os.environ['_DX_DUMP_BILLED_PROJECT'], "w") as fd:
                        if project is not None and project != DXFile.NO_PROJECT_HINT:
                            fd.write(project)
                # The idea here is to cache a download URL for the entire file, that will
                # be good for a few minutes. This avoids each thread having to ask the
                # server for a URL, increasing server load.
                #
                # To avoid thread race conditions, this check/update procedure is protected
                # with a lock.

                # logging.debug("Download URL unset or expired, requesting a new one")
                if "timeout" not in kwargs:
                    kwargs["timeout"] = FILE_REQUEST_TIMEOUT
                resp = dxpy.api.file_download(self._dxid, args, **kwargs)
                self._download_url = resp["url"]
                self._download_url_headers = _validate_headers(resp.get("headers", {}))
                if preauthenticated:
                    self._download_url_expires = resp["expires"]/1000 - 60  # Try to account for drift
                else:
                    self._download_url_expires = 32503680000  # doesn't expire (year 3000)

            # Make a copy, ensuring each thread has its own mutable
            # version of the headers.  Note: python strings are
            # immutable, so we can safely give a reference to the
            # download url.
            retval_download_url = self._download_url
            retval_download_url_headers = copy.copy(self._download_url_headers)

        return retval_download_url, retval_download_url_headers

    def _generate_read_requests(self, start_pos=0, end_pos=None, project=None,
                                limit_chunk_size=None, **kwargs):
        # project=None means no hint is to be supplied to the apiserver. It is
        # an error to supply a project that does not contain this file.
        if limit_chunk_size is None:
            limit_chunk_size = self._read_bufsize

        if self._file_length == None:
            desc = self.describe(**kwargs)
            self._file_length = int(desc["size"])

        if end_pos == None:
            end_pos = self._file_length
        if end_pos > self._file_length:
            raise DXFileError("Invalid end_pos")

        def chunk_ranges(start_pos, end_pos, init_chunk_size=1024*64, ramp=2, num_requests_between_ramp=4):
            cur_chunk_start = start_pos
            cur_chunk_size = min(init_chunk_size, limit_chunk_size)
            i = 0
            while cur_chunk_start < end_pos:
                cur_chunk_end = min(cur_chunk_start + cur_chunk_size - 1, end_pos)
                yield cur_chunk_start, cur_chunk_end
                cur_chunk_start += cur_chunk_size
                if cur_chunk_size < limit_chunk_size and i % num_requests_between_ramp == (num_requests_between_ramp - 1):
                    cur_chunk_size = min(cur_chunk_size * ramp, limit_chunk_size)
                i += 1

        for chunk_start_pos, chunk_end_pos in chunk_ranges(start_pos, end_pos):
            url, headers = self.get_download_url(project=project, **kwargs)
            # It is possible for chunk_end_pos to be outside of the range of the file
            yield dxpy._dxhttp_read_range, [url, headers, chunk_start_pos, min(chunk_end_pos, self._file_length - 1),
                                            FILE_REQUEST_TIMEOUT], {}

    def _next_response_content(self, get_first_chunk_sequentially=False):
        if self._response_iterator is None:
            self._response_iterator = dxpy.utils.response_iterator(
                self._request_iterator,
                self._http_threadpool,
                do_first_task_sequentially=get_first_chunk_sequentially
            )
        try:
            return next(self._response_iterator)
        except:
            # If an exception is raised, the iterator is unusable for
            # retrieving any more items. Destroy it so we'll reinitialize it
            # next time.
            self._response_iterator = None
            self._request_iterator = None
            raise

    def _read2(self, length=None, use_compression=None, project=None, **kwargs):
        '''
        :param length: Maximum number of bytes to be read
        :type length: integer
        :param project: project to use as context for this download (may affect
            which billing account is billed for this download). If specified,
            must be a project in which this file exists. If not specified, the
            project ID specified in the handler is used for the download, IF it
            contains this file. If set to DXFile.NO_PROJECT_HINT, no project ID
            is supplied for the download, even if the handler specifies a
            project ID.
        :type project: str or None
        :rtype: string
        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if *project* is supplied
           and it does not contain this file

        Returns the next *length* bytes, or all the bytes until the end of file
        (if no *length* is given or there are fewer than *length* bytes left in
        the file).

        .. note:: After the first call to read(), the project arg and
           passthrough kwargs are not respected while using the same response
           iterator (i.e. until next seek).

        '''
        if self._file_length == None:
            desc = self.describe(**kwargs)
            if desc["state"] != "closed":
                raise DXFileError("Cannot read from file until it is in the closed state")
            self._file_length = int(desc["size"])

        # If running on a worker, wait for the first file download chunk
        # to come back before issuing any more requests. This ensures
        # that all subsequent requests can take advantage of caching,
        # rather than having all of the first DXFILE_HTTP_THREADS
        # requests simultaneously hit a cold cache. Enforce a minimum
        # size for this heuristic so we don't incur the overhead for
        # tiny files (which wouldn't contribute as much to the load
        # anyway).
        get_first_chunk_sequentially = (self._file_length > 128 * 1024 and self._pos == 0 and dxpy.JOB_ID)

        if self._pos == self._file_length:
            return b""

        if length == None or length > self._file_length - self._pos:
            length = self._file_length - self._pos

        buf = self._read_buf
        buf_remaining_bytes = dxpy.utils.string_buffer_length(buf) - buf.tell()
        if length <= buf_remaining_bytes:
            self._pos += length
            return buf.read(length)
        else:
            orig_buf_pos = buf.tell()
            orig_file_pos = self._pos
            buf.seek(0, os.SEEK_END)
            self._pos += buf_remaining_bytes
            while self._pos < orig_file_pos + length:
                remaining_len = orig_file_pos + length - self._pos

                if self._response_iterator is None:
                    self._request_iterator = self._generate_read_requests(
                        start_pos=self._pos, project=project, **kwargs)

                content = self._next_response_content(get_first_chunk_sequentially=get_first_chunk_sequentially)

                if len(content) < remaining_len:
                    buf.write(content)
                    self._pos += len(content)
                else: # response goes beyond requested length
                    buf.write(content[:remaining_len])
                    self._pos += remaining_len
                    self._read_buf = BytesIO()
                    self._read_buf.write(content[remaining_len:])
                    self._read_buf.seek(0)
            buf.seek(orig_buf_pos)
            return buf.read()

    def read(self, length=None, use_compression=None, project=None, **kwargs):
        # Check if the file is present in dxfile project attribute if the project arg not specified 
        if project is None and self._exists_in_proj is None and self.get_proj_id() is not None:
            self._exists_in_proj = object_exists_in_project(self.get_id(), self.get_proj_id())
        # Use the DXFile attribute if the project arg is not provided
        if project is None and self._exists_in_proj:
            project = self.get_proj_id()
        data = self._read2(length=length, use_compression=use_compression, project=project, **kwargs)
        # In python3, the underlying system methods use the 'bytes' type, not 'string'
        if self._binary_mode is True:
            return data
        return data.decode("utf-8")
    
    def archive(self, all_copies=False):
        '''
        :param all_copies: Force the transition of files into the archived state. Requesting user must be the ADMIN of the project billTo org. 
            If true, archive all the copies of files in projects with the same billTo org.
        :type all_copies: boolean
        :raises: :exc:`~dxpy.exceptions.InvalidState` if the file is not in a live state
        :raises: :exc:`~dxpy.exceptions.PermissionDenied` if the requesting user does not have CONTRIBUTE access or
            is not an ADMIN of the project billTo org with allCopies=True.
        '''
        dxpy.api.project_archive(self.get_proj_id(), {"files": [self.get_id()], "allCopies": all_copies})

    def unarchive(self, dry_run=False):
        '''
        :param dry_run:  If true, only display the output of the API call without executing the unarchival
        :type dry_run: boolean
        :raises: :exc:`~dxpy.exceptions.InvalidState` if the file is not in a closed or archived state
        :raises: :exc:`~dxpy.exceptions.PermissionDenied` if the requesting user does not have CONTRIBUTE access    
        '''
        dxpy.api.project_unarchive(self.get_proj_id(), {"files": [self.get_id()], "dryRun": dry_run})



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import re

maketrans = bytes.maketrans

COMPLEMENT = maketrans(b"ATGCatgc", b"TACGTACG")

SEQ_PATTERN = re.compile(b'[ACGTacgtNn]*$')

def reverse_complement(seq):
    if isinstance(seq, str):
        bytes_seq = seq.encode('utf-8')
    else:
        bytes_seq = seq
    if not SEQ_PATTERN.match(bytes_seq):
        raise ValueError('Sequence %r must consist only of A, C, G, T, N' % (seq,))
    return bytes_seq.translate(COMPLEMENT)[::-1]



#!/usr/bin/env python3
#
# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


'''
Executable Builder
+++++++++++++++++++

Contains utility methods useful for deploying executables (apps, applets, workflows)
onto the platform.

'''

from __future__ import print_function, unicode_literals, division, absolute_import
import os
import re
import collections

from .utils.resolver import resolve_path, is_container_id
from .cli import try_call
import dxpy

GLOBAL_EXEC_NAME_RE = re.compile("^[a-zA-Z0-9._\-]+$")
GLOBAL_EXEC_VERSION_RE = re.compile("^([1-9][0-9]*|0)\.([1-9][0-9]*|0)\.([1-9][0-9]*|0)(-[-0-9A-Za-z]+(\.[-0-9A-Za-z]+)*)?(\+[-0-9A-Za-z]+(\.[-0-9A-Za-z]+)*)?$")

def get_parsed_destination(dest_str):
    """
    Parses dest_str, which is (roughly) of the form
    PROJECT:/FOLDER/NAME, and returns a tuple (project, folder, name)
    """
    # Interpret strings of form "project-XXXX" (no colon) as project. If
    # we pass these through to resolve_path they would get interpreted
    # as folder names...
    if is_container_id(dest_str):
        return (dest_str, None, None)

    # ...otherwise, defer to resolver.resolve_path. This handles the
    # following forms:
    #
    # /FOLDER/
    # /ENTITYNAME
    # /FOLDER/ENTITYNAME
    # [PROJECT]:
    # [PROJECT]:/FOLDER/
    # [PROJECT]:/ENTITYNAME
    # [PROJECT]:/FOLDER/ENTITYNAME
    return try_call(resolve_path, dest_str)


def inline_documentation_files(json_spec, src_dir):
    """
    Modifies the provided json_spec dict (which may be an app, applet,
    workflow spec) to inline the contents of the readme file into
    "description" and the developer readme into "developerNotes".
    """
    # Inline description from a readme file
    if 'description' not in json_spec:
        readme_filename = None
        for filename in 'README.md', 'Readme.md', 'readme.md':
            if os.path.exists(os.path.join(src_dir, filename)):
                readme_filename = filename
                break
        if readme_filename is not None:
            with open(os.path.join(src_dir, readme_filename)) as fh:
                json_spec['description'] = fh.read()

    # Inline developerNotes from Readme.developer.md
    if 'developerNotes' not in json_spec:
        for filename in 'README.developer.md', 'Readme.developer.md', 'readme.developer.md':
            if os.path.exists(os.path.join(src_dir, filename)):
                with open(os.path.join(src_dir, filename)) as fh:
                    json_spec['developerNotes'] = fh.read()
                break


def delete_temporary_projects(projects):
    """
    Destroys all projects from the list.
    """
    for project in projects:
        try:
            dxpy.api.project_destroy(project)
        except Exception:
            pass

def get_valid_bill_to(bill_to, executable_builder_exception):
    """
    Check if the requesting user can perform billable activities on behalf of the billTo
    If not specified, default to the billTo of the requesting user
    otherwise it must be either the ID of the requesting user, 
    or an org of which the requesting user is a member with 'allowBillableActivities' permission
    """
    user_id = dxpy.whoami()
    if not bill_to:
        return dxpy.api.user_describe(user_id)['billTo']

    exception_msg = None
    if bill_to.startswith('user-') and bill_to != user_id:
        exception_msg = 'Cannot request another user to be the "billTo"'
    elif bill_to.startswith('org-'):
        try:
            member_access = dxpy.api.org_describe(bill_to)
            if not member_access['allowBillableActivities']:
                exception_msg='You are not a member in {} with allowBillableActivities permission. Please check the billing policy of the org.'.format(bill_to)
        except:
            exception_msg='Cannot retrieve billing information for {}. Please check your access level and the billing policy of the org.'.format(bill_to)
    else:
        exception_msg='The field "billTo" must be a valid ID of a user/org.'
    
    if exception_msg:
        raise executable_builder_exception(exception_msg)
    
    return bill_to

def verify_developer_rights(prefixed_name):
    """
    Checks if the current user is a developer of the app or global workflow
    with the given name. If the app/global workflow exists and the user has
    developer rights to it, the function returns a named tuple representing
    the executable that was queried.
    """
    assert(prefixed_name.startswith('app-') or prefixed_name.startswith('globalworkflow-'))

    if prefixed_name.partition('-')[0] == 'app':
        exception_type = dxpy.app_builder.AppBuilderException
        describe_method = dxpy.api.app_describe
        exception_msg = \
            'An app with the given name already exists and you are not a developer of that app'
    else:
        exception_type = dxpy.workflow_builder.WorkflowBuilderException
        describe_method = dxpy.api.global_workflow_describe
        exception_msg = \
            'A global workflow with the given name already exists and you are not a developer of that workflow'
    
    name_already_exists = True
    is_developer = False
    version = None
    executable_id = None
    FoundExecutable = collections.namedtuple('FoundExecutable', ['name', 'version', 'id'])
    try:
        describe_output = describe_method(prefixed_name,
                                          input_params={"fields": {"isDeveloperFor": True,
                                                                   "version": True,
                                                                   "id": True}})
        is_developer = describe_output['isDeveloperFor']
        version = describe_output['version']
        executable_id = describe_output['id']
    except dxpy.exceptions.DXAPIError as e:
        if e.name == 'ResourceNotFound':
            name_already_exists = False
        elif e.name == 'PermissionDenied':
            raise exception_type(exception_msg)
        else:
            raise e

    if not name_already_exists:
        # This app/workflow doesn't exist yet so its creation will succeed
        # (or at least, not fail on the basis of the ACL).
        return FoundExecutable(name=None, version=None, id=None)

    name_without_prefix = prefixed_name.partition('-')[2]
    if not is_developer:
        raise exception_type('You are not a developer for {n}'.format(n=name_without_prefix))

    return FoundExecutable(name=name_without_prefix, version=version, id=executable_id)


def assert_consistent_regions(from_spec, from_command_line, executable_builder_exception):
    """
    Verifies the regions passed with --region CLI argument and the ones
    specified in regionalOptions are the same (if both CLI and spec were used)
    """
    if from_spec is None or from_command_line is None:
        return
    if set(from_spec) != set(from_command_line):
        raise executable_builder_exception("--region and the 'regionalOptions' key in the JSON file or --extra-args do not agree")


def assert_consistent_reg_options(exec_type, json_spec, executable_builder_exception):
    """
    Validates the "regionalOptions" field and verifies all the regions used
    in "regionalOptions" have the same options.
    """
    reg_options_spec = json_spec.get('regionalOptions')
    json_fn = 'dxapp.json or --extra-args' if exec_type == 'app' else 'dxworkflow.json or --extra-args'

    if not isinstance(reg_options_spec, dict):
        raise executable_builder_exception("The field 'regionalOptions' in  must be a mapping")
    if not reg_options_spec:
        raise executable_builder_exception(
            "The field 'regionalOptions' in " + json_fn + " must be a non-empty mapping")
    regional_options_list = list(reg_options_spec.items())
    for region, opts_for_region in regional_options_list:
        if not isinstance(opts_for_region, dict):
            raise executable_builder_exception("The field 'regionalOptions['" + region +
                            "']' in " + json_fn + " must be a mapping")
        if set(opts_for_region.keys()) != set(regional_options_list[0][1].keys()):
            if set(opts_for_region.keys()) - set(regional_options_list[0][1].keys()):
                with_key, without_key = region, regional_options_list[0][0]
                key_name = next(iter(set(opts_for_region.keys()) - set(regional_options_list[0][1].keys())))
            else:
                with_key, without_key = regional_options_list[0][0], region
                key_name = next(iter(set(regional_options_list[0][1].keys()) - set(opts_for_region.keys())))
            raise executable_builder_exception(
                "All regions in regionalOptions must specify the same options; " +
                "{} was given for {} but not for {}" .format (key_name, with_key, without_key)
            )

        if exec_type == 'app':
            for key in opts_for_region:
                if key in json_spec.get('runSpec', {}):
                    raise executable_builder_exception(
                    key + " cannot be given in both runSpec and in regional options for " + region)

def get_permitted_regions(bill_to, executable_builder_exception):
    """
    Validates requested bill_to and returns the set of its permitted regions.
    """
    billable_regions = set()
    try:
        billable_regions= set(dxpy.DXHTTPRequest('/' + bill_to + '/describe', {}).get("permittedRegions"))
    except:
        raise executable_builder_exception("Failed to get permitted regions of {}".format(bill_to))
    return billable_regions

def get_enabled_regions(exec_type, json_spec, from_command_line, executable_builder_exception):
    """
    Return a list of regions in which the global executable (app or global workflow)
    will be enabled, based on the "regionalOption" in their JSON specification
    and/or --region CLI argument used with "dx build".

    :param exec_type: 'app' or 'globalworkflow'
    :type json_spec: str.
    :param json_spec: The contents of dxapp.json or dxworkflow.json
    :type json_spec: dict or None.
    :param from_command_line: The regional options specified on the command-line via --region.
    :type from_command_line: list or None.
    :param executable_builder_exception: Exception that will be thrown.
    :type executable_builder_exception: AppBuilderException or WorkflowBuilderException.
    """

    from_spec = json_spec.get('regionalOptions')

    if from_spec is not None:
        assert_consistent_reg_options(exec_type, json_spec, executable_builder_exception)

    assert_consistent_regions(from_spec, from_command_line, executable_builder_exception)

    enabled_regions = None
    if from_spec is not None:
        enabled_regions = from_spec.keys()
    elif from_command_line is not None:
        enabled_regions = from_command_line

    return enabled_regions



#!/usr/bin/env python

import os
import subprocess

from dxpy import DXFile, config, upload_local_file
from dxpy.exceptions import err_exit


class ImageRef(object):
    def __init__(
            self,
            process,
            digest,
            dx_file_id=None,
            repository=None,
            image_name=None,
            tag=None
    ):
        """
        A class to handle an image reference from nextflow pipeline.
        :param process: An NPA proces name (aka task name) which uses a given image
        :type process: String
        :param digest: An image digest
        :type digest: String
        :param dx_file_id: dx file id on the platform
        :type dx_file_id: Optional[String]
        :param repository: Image repository
        :type repository: Optional[String]
        :param image_name: Image name (usually a basename of the image referenced with repository)
        :type image_name: Optional[String]
        :param tag: A version tag
        :type tag: Optional[String]
        """
        self._caching_dir = os.path.join("/.cached_docker_images/", image_name or "")
        self._dx_file_id = dx_file_id
        self._bundled_depends = None
        self._repository = repository
        self._image_name = image_name
        self._tag = tag
        self._digest = digest
        self._process = process

    @property
    def bundled_depends(self):
        if not self._bundled_depends:
            self._bundled_depends = self._package_bundle()
        return self._bundled_depends

    @property
    def identifier(self):
        return self._join_if_exists("_", [self._repository, self._image_name, self._tag, self._digest])

    def _cache(self, file_name):
        """
        Function to store an image on the platform as a dx file object. Should be implemented in subclasses.
        :param file_name: A file name under which the image will be saved on the platform
        :type file_name: String
        :returns: Tuple[String, String] dx file id, file name (basename)
        """
        raise NotImplementedError("Abstract class. Method not implemented. Use the concrete implementations.")

    def _reconstruct_image_ref(self):
        raise NotImplementedError("Abstract class. Method not implemented. Use the concrete implementations.")

    def _construct_cache_file_name(self):
        return self._join_if_exists("_", [self._image_name, self._tag])

    @staticmethod
    def _join_if_exists(delimiter, parts):
        return delimiter.join([x for x in parts if x])

    def _dx_file_get_name(self):
        dx_file_handle = DXFile(self._dx_file_id, config["DX_PROJECT_CONTEXT_ID"])
        return dx_file_handle.describe().get("name")

    def _package_bundle(self):
        """
        Function to include a container image stored on the platform into NPA
        :returns: Dict in the format of {"name": "bundle.tar.gz", "id": {"$dnanexus_link": "file-xxxx"}}
        """
        if not self._dx_file_id:
            cache_file_name = self._construct_cache_file_name()
            self._dx_file_id = self._cache(cache_file_name)
        else:
            cache_file_name = self._dx_file_get_name()
        return {
            "name": cache_file_name,
            "id": {"$dnanexus_link": self._dx_file_id}
        }


class DockerImageRef(ImageRef):
    def __init__(
            self,
            process,
            digest,
            dx_file_id=None,
            repository=None,
            image_name=None,
            tag=None
    ):
        super().__init__(
            process,
            digest,
            dx_file_id,
            repository,
            image_name,
            tag)

    def _cache(self, file_name):
        full_image_ref = self._reconstruct_image_ref()
        docker_pull_cmd = "sudo docker pull {}".format(full_image_ref)
        docker_save_cmd = "sudo docker save {} | gzip > {}".format(full_image_ref, file_name)
        for cmd in [docker_pull_cmd, docker_save_cmd]:
            try:
                _ = subprocess.check_output(cmd, shell=True)
            except subprocess.CalledProcessError:
                err_exit("Failed to run a subprocess command: {}".format(cmd))
        # may need wait_on_close = True??
        extracted_digest = self._digest
        if not self._digest:
            digest_cmd = "docker images --no-trunc --quiet {}".format(full_image_ref)
            extracted_digest = subprocess.check_output(digest_cmd, shell=True).decode().strip()
        uploaded_dx_file = upload_local_file(
            filename=file_name,
            project=config["DX_PROJECT_CONTEXT_ID"],
            folder=self._caching_dir,
            name=file_name,
            parents=True,
            properties={"image_digest": self._digest or extracted_digest}
        )
        return uploaded_dx_file.get_id()

    def _reconstruct_image_ref(self):
        """
        Docker image reference has the form of <REPOSITORY_NAME>/<IMAGE_NAME>:<VERSION_TAG> or
        <REPOSITORY_NAME>/<IMAGE_NAME>@<DIGEST>
        """
        repo_and_image_name = self._join_if_exists("", [self._repository, self._image_name])
        if self._digest:
            full_ref = self._join_if_exists("@", [repo_and_image_name, self._digest])
        else:
            full_ref = self._join_if_exists(":", [repo_and_image_name, self._tag])
        return full_ref



import sys
from ..exceptions import err_exit

# Generic error messages
malformed_filter = 'Found following invalid filters: {}'
maxitem_message = 'Too many items given in field {}, maximum is {}'
# An integer equel to 2 if script is run with python2, and 3 if run with python3
python_version = sys.version_info.major

def is_list_of_strings(object):
    if not isinstance(object, list):
        return False
    for item in object:
        # Note that in python 2.7 these strings are read in as unicode
        if python_version == 2:
            if not (isinstance(item, str) or isinstance(item, unicode)):
                return False
        else:
            if not isinstance(item, str):
                return False
    return True

def validate_somatic_filter(filter, filter_type):
    keys = filter.keys()
    if filter_type == 'variant':
        required_filter_count = 0
        expected_keys = {'annotation', 'allele', 'sample', 'location'}
        if not set(keys).issubset(expected_keys):
            err_exit(malformed_filter.format(str(set(keys).difference(list(expected_keys)))))

        if 'annotation' in keys:
            annotation_filter = filter['annotation']
            sub_keys = annotation_filter.keys()
            expected_sub_keys = {'symbol', 'gene', 'feature', 'hgvsc', 'hgvsp'}
            if not set(sub_keys).issubset(expected_sub_keys):
                err_exit(malformed_filter.format(str(set(sub_keys).difference(list(expected_sub_keys)))))

            if 'symbol' in sub_keys:
                if not is_list_of_strings(annotation_filter['symbol']):
                    err_exit(malformed_filter.format('annotation["symbol"]'))
                if len(annotation_filter['symbol']) > 100:
                    err_exit(maxitem_message.format('annotation["symbol"]',100))
                if annotation_filter['symbol']:
                    required_filter_count += 1

            if 'gene' in sub_keys:
                if not is_list_of_strings(annotation_filter['gene']):
                    err_exit(malformed_filter.format('annotation["gene"]'))
                if len(annotation_filter['gene']) > 100:
                    err_exit(maxitem_message.format('annotation["gene"]',100))
                if annotation_filter['gene']:
                    required_filter_count += 1

            if 'feature' in sub_keys:
                if not is_list_of_strings(annotation_filter['feature']):
                    err_exit(malformed_filter.format('annotation["feature"]'))
                if len(annotation_filter['feature']) > 100:
                    err_exit(maxitem_message.format('annotation["feature"]',100))
                if annotation_filter['feature']:
                    required_filter_count += 1

            if 'hgvsc' in sub_keys:
                if not is_list_of_strings(annotation_filter['hgvsc']):
                    err_exit(malformed_filter.format('annotation["hgvsc"]'))
                if len(annotation_filter['hgvsc']) > 100:
                    err_exit(maxitem_message.format('annotation["hgvsc"]',100))
            
            if 'hgvsp' in sub_keys:
                if not is_list_of_strings(annotation_filter['hgvsp']):
                    err_exit(malformed_filter.format('annotation["hgvsp"]'))
                if len(annotation_filter['hgvsp']) > 100:
                    err_exit(maxitem_message.format('annotation["hgvsp"]',100))
        
        if 'allele' in keys:
            allele_filter = filter['allele']
            sub_keys = allele_filter.keys()
            expected_sub_keys = {'allele_id', 'variant_type'}
            if not set(sub_keys).issubset(expected_sub_keys):
                err_exit(malformed_filter.format(str(set(sub_keys).difference(list(expected_sub_keys)))))

            if 'allele_id' in sub_keys:
                if not is_list_of_strings(allele_filter['allele_id']):
                    err_exit(malformed_filter.format('allele["allele_id"]'))
                if len(allele_filter['allele_id']) > 100:
                    err_exit(maxitem_message.format('allele["allele_id"]',100))
                if allele_filter['allele_id']:
                    required_filter_count += 1

            if 'variant_type' in sub_keys:
                if not is_list_of_strings(allele_filter['variant_type']):
                    err_exit(malformed_filter.format('allele["variant_type"]'))
                for item in allele_filter['variant_type']:
                    if item not in ['SNP', 'INS', 'DEL', 'DUP', 'INV', 'CNV', 'CNV:TR', 'BND', 'DUP:TANDEM', 'DEL:ME', 'INS:ME', 'MISSING', 'MISSING:DEL', 'UNSPECIFIED', 'REF', 'OTHER']:
                        err_exit(malformed_filter.format('allele["variant_type"]'))
                if len(allele_filter['variant_type']) > 10:
                    err_exit(maxitem_message.format('allele["variant_type"]',10))

        if 'sample' in keys:
            sample_filter = filter['sample']
            sub_keys = sample_filter.keys()
            expected_sub_keys = {'sample_id', 'assay_sample_id'}
            if not set(sub_keys).issubset(expected_sub_keys):
                err_exit(malformed_filter.format(str(set(sub_keys).difference(list(expected_sub_keys)))))

            if 'sample_id' in sub_keys:
                if not is_list_of_strings(sample_filter['sample_id']):
                    err_exit(malformed_filter.format('sample["sample_id"]'))
                if len(sample_filter['sample_id']) > 500:
                    err_exit(maxitem_message.format('sample["sample_id"]',500))
            
            if 'assay_sample_id' in sub_keys:
                if not is_list_of_strings(sample_filter['assay_sample_id']):
                    err_exit(malformed_filter.format('sample["assay_sample_id"]'))
                if len(sample_filter['assay_sample_id']) > 1000:
                    err_exit(maxitem_message.format('sample["assay_sample_id"]',1000))
                   
        if 'location' in keys:
            # Ensure there are not more than 100 locations
            if len(filter['location']) > 100:
                print(maxitem_message.format('location',100))
            for indiv_location in filter['location']:
                indiv_loc_keys = indiv_location.keys()
                # Ensure all keys are there
                if not (
                    ('chromosome' in indiv_loc_keys)
                    and ('starting_position' in indiv_loc_keys)
                    and ('ending_position' in indiv_loc_keys)
                ):
                    print(malformed_filter.format('location'))
                    err_exit()
                # Check that each key is a string
                if not is_list_of_strings(list(indiv_location.values())):
                    print(malformed_filter.format('location'))
                    err_exit()
            if filter['location']:
                required_filter_count += 1

    if required_filter_count != 1:
        err_exit('Exactly one of "symbol", "gene", feature", "allele_id" or "location" must be provided in the json')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
from __future__ import print_function, unicode_literals, division, absolute_import

import os
import sys
import unittest

from parameterized import parameterized
from dxpy_testutil import DXTestCase
from dxpy.nextflow.ImageRefFactory import ImageRefFactory, ImageRefFactoryError
from dxpy.nextflow.ImageRef import DockerImageRef

spawn_extra_args = {"encoding": "utf-8"}


class TestImageRef(DXTestCase):
    @parameterized.expand([
        [{"engine": "docker", "process": "proc1", "digest": "sha256aasdfadfadfafddasfdsfa"}]
    ])
    def test_ImageRefFactory(self, image_ref):
        image_ref_factory = ImageRefFactory(image_ref)
        image = image_ref_factory.get_image()
        self.assertTrue(isinstance(image, DockerImageRef))

    @parameterized.expand([
        [{"process": "proc1", "digest": "sha256aasdfadfadfafddasfdsfa"}, "Provide the container engine"],
        [{"engine": "singularity", "process": "proc1", "digest": "sha256aasdfadfadfafddasfdsfa"}, "Unsupported container engine: singularity"]
    ])
    def test_ImageRefFactory_errors(self, image_ref, exception):
        with self.assertRaises(ImageRefFactoryError) as err:
            image_ref_factory = ImageRefFactory(image_ref)
            _ = image_ref_factory.get_image()
            self.assertEqual(err.exception, exception)



if __name__ == '__main__':
    if 'DXTEST_FULL' not in os.environ:
        sys.stderr.write(
            'WARNING: env var DXTEST_FULL is not set; tests that create apps or run jobs will not be run\n')
    unittest.main()



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import print_function, unicode_literals, division, absolute_import

import socket, json, time, os, logging

from logging.handlers import SysLogHandler

from dxpy.exceptions import DXError

class DXLogHandler(SysLogHandler):
    '''
    Logging handler for DNAnexus application level logging.
    Code adapted from logging.handlers.SysLogHandler.

    This handler is automatically enabled in the job template when running Python code in the execution environment.
    It forwards all log messages sent through the :mod:`logging` module to the DNAnexus log service,
    so that they can be examined through the log query API.
    To enable the handler in a Python subprocess in the execution environment, use:

        import logging
        logging.basicConfig(level=logging.DEBUG)
        from dxpy.dxlog import DXLogHandler
        logging.getLogger().addHandler(DXLogHandler())

    '''
    def __init__(self, priority_log_address="/opt/dnanexus/log/priority",
                 bulk_log_address="/opt/dnanexus/log/bulk",
                 source="DX_APP"):
        logging.Handler.__init__(self)

        self.priority_log_address = priority_log_address
        self.priority_log_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        self.bulk_log_address = bulk_log_address
        self.bulk_log_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        if not os.path.exists(priority_log_address):
            raise DXError("The path %s does not exist, but is required for application logging" % (priority_log_address))
        if not os.path.exists(bulk_log_address):
            raise DXError("The path %s does not exist, but is required for application logging" % (bulk_log_address))

        self.priority_log_socket.connect(priority_log_address)
        self.bulk_log_socket.connect(bulk_log_address)

        self.source = source

    def close(self):
        self.priority_log_socket.close()
        self.bulk_log_socket.close()
        logging.Handler.close(self)

    def encodePriority(self, record):
        # See logging.handlers.SysLogHandler for an explanation of this.
        return self.priority_names[self.priority_map.get(record.levelname, "warning")]

    def truncate_message(self, message):
        msg_bytes = message.encode('utf-8')

        if len(json.dumps(message)) <= 8015:
            return message

        msg_bytes = msg_bytes[:8000]
        while len(json.dumps(_bytes2utf8(msg_bytes))) > 8000:
            msg_bytes = msg_bytes[:-1]

        message = _bytes2utf8(msg_bytes)
        return message + "... [truncated]"

    def is_resource_log(self, message):
        return message.startswith("CPU: ")

    def emit(self, record):
        level = self.encodePriority(record)
        message = record.getMessage()
        # The Linux domain socket datagram size limit is 8 KB, but
        # with the extra padding introduced by the log function, the
        # incoming message needs to be smaller - we truncate it to
        # at most 8015 bytes here.
        message = self.truncate_message(message)

        data = json.dumps({"source": self.source, "timestamp": int(round(time.time() * 1000)),
                           "level": level, "msg": message}).encode('utf-8', 'ignore')

        levelno = int(record.levelno)
        if levelno >= logging.CRITICAL or (levelno == logging.INFO and self.is_resource_log(message)):
            # Critical, alert, emerg, or resource status
            cur_socket = self.priority_log_socket
            cur_socket_address = self.priority_log_address
        else:
            cur_socket = self.bulk_log_socket
            cur_socket_address = self.bulk_log_address

        try:
            cur_socket.send(data)
        except socket.error:
            cur_socket.connect(cur_socket_address)
            cur_socket.send(data)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

def _bytes2utf8(bytes):
    """
    Convert bytes to a UTF-8 string and ignore UnicodeDecodeError for chars that could have been messed up by truncating.
    """
    return bytes.decode('utf-8', 'ignore')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
from __future__ import print_function, unicode_literals, division, absolute_import

import unittest
import tempfile
import shutil
import os
import re
import subprocess
import pandas as pd
import dxpy
from dxpy_testutil import cd, chdir

dirname = os.path.dirname(__file__)

class TestDXExtractDataset(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_file_dir = os.path.join(dirname, "extract_dataset_test_files")
        proj_name = "dx-toolkit_test_data"
        proj_id = list(dxpy.find_projects(describe=False, level='VIEW', name=proj_name))[0]['id']
        cd(proj_id + ":/")

    def test_e2e_dataset_ddd(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        out_directory = tempfile.mkdtemp()
        cmd = ["dx", "extract_dataset", dataset_record, "-ddd", "-o", out_directory]
        subprocess.check_call(cmd)
        self.end_to_end_ddd(out_directory=out_directory, rec_name = "extract_dataset_test")

    def test_e2e_cohortbrowser_ddd(self):
        cohort_record = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test"
        out_directory = tempfile.mkdtemp()
        cmd = ["dx", "extract_dataset", cohort_record, "-ddd", "-o", out_directory]
        subprocess.check_call(cmd)
        self.end_to_end_ddd(out_directory=out_directory, rec_name = "Combined_Cohort_Test")

    def test_e2e_dataset_sql(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        truth_output = "SELECT `patient_0001_1`.`patient_id` AS `patient.patient_id`, `patient_0001_1`.`name` AS `patient.name`, `patient_0002_1`.`weight` AS `patient.weight`, `patient_0001_1`.`date_of_birth` AS `patient.date_of_birth`, `patient_0002_1`.`verified_dtm` AS `patient.verified_dtm`, `test_1`.`test_id` AS `test.test_id`, `trial_visit_0001_1`.`visit_id` AS `trial_visit.visit_id`, `baseline_0001_1`.`baseline_id` AS `baseline.baseline_id`, `hospital_0001_1`.`hospital_id` AS `hospital.hospital_id`, `doctor_0001_1`.`doctor_id` AS `doctor.doctor_id` FROM `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0001` AS `patient_0001_1` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0002` AS `patient_0002_1` ON `patient_0001_1`.`patient_id` = `patient_0002_1`.`patient_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`trial_visit_0001` AS `trial_visit_0001_1` ON `patient_0001_1`.`patient_id` = `trial_visit_0001_1`.`visit_patient_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`test` AS `test_1` ON `trial_visit_0001_1`.`visit_id` = `test_1`.`test_visit_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`baseline_0001` AS `baseline_0001_1` ON `patient_0001_1`.`patient_id` = `baseline_0001_1`.`b_patient_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`hospital_0001` AS `hospital_0001_1` ON `patient_0001_1`.`hid` = `hospital_0001_1`.`hospital_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`doctor_0001` AS `doctor_0001_1` ON `trial_visit_0001_1`.`visit_did` = `doctor_0001_1`.`doctor_id`;"
        cmd = ["dx", "extract_dataset", dataset_record, "--fields", "patient.patient_id, patient.name,patient.weight, patient.date_of_birth, patient.verified_dtm, test.test_id, trial_visit.visit_id, baseline.baseline_id, hospital.hospital_id , doctor.doctor_id",
               "--sql", "-o", "-"]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        stdout = process.communicate()[0]
        self.assertTrue(re.match(truth_output,stdout.strip()))

    def test_e2e_cohortbrowser_sql(self):
        cohort_record = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test"
        truth_output = "SELECT `patient_0001_1`.`patient_id` AS `patient.patient_id`, `patient_0001_1`.`name` AS `patient.name`, `patient_0002_1`.`weight` AS `patient.weight`, `patient_0001_1`.`date_of_birth` AS `patient.date_of_birth`, `patient_0002_1`.`verified_dtm` AS `patient.verified_dtm`, `test_1`.`test_id` AS `test.test_id`, `trial_visit_0001_1`.`visit_id` AS `trial_visit.visit_id`, `baseline_0001_1`.`baseline_id` AS `baseline.baseline_id`, `hospital_0001_1`.`hospital_id` AS `hospital.hospital_id`, `doctor_0001_1`.`doctor_id` AS `doctor.doctor_id` FROM `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0001` AS `patient_0001_1` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0002` AS `patient_0002_1` ON `patient_0001_1`.`patient_id` = `patient_0002_1`.`patient_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`trial_visit_0001` AS `trial_visit_0001_1` ON `patient_0001_1`.`patient_id` = `trial_visit_0001_1`.`visit_patient_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`test` AS `test_1` ON `trial_visit_0001_1`.`visit_id` = `test_1`.`test_visit_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`baseline_0001` AS `baseline_0001_1` ON `patient_0001_1`.`patient_id` = `baseline_0001_1`.`b_patient_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`hospital_0001` AS `hospital_0001_1` ON `patient_0001_1`.`hid` = `hospital_0001_1`.`hospital_id` LEFT OUTER JOIN `database_[a-z0-9]{24}__extract_dataset_test`.`doctor_0001` AS `doctor_0001_1` ON `trial_visit_0001_1`.`visit_did` = `doctor_0001_1`.`doctor_id` WHERE `patient_0001_1`.`patient_id` IN \(SELECT `cohort_query`.`patient_id` FROM \(SELECT `patient_0001_1`.`patient_id` AS `patient_id` FROM `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0001` AS `patient_0001_1` WHERE `patient_0001_1`.`patient_id` IN \(SELECT `patient_id` FROM \(SELECT DISTINCT `cohort_subquery`.`patient_id` AS `patient_id` FROM \(SELECT DISTINCT `patient_0001_1`.`patient_id` AS `patient_id`, `patient_0001_1`.`hid` AS `hid` FROM `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0001` AS `patient_0001_1` WHERE EXISTS \(SELECT `hospital_0001_1`.`hospital_id` AS `hospital_id` FROM `database_[a-z0-9]{24}__extract_dataset_test`.`hospital_0001` AS `hospital_0001_1` WHERE `hospital_0001_1`.`hospital_id` BETWEEN 2 AND 5 AND `hospital_0001_1`.`hospital_id` = `patient_0001_1`.`hid`\)\) AS `cohort_subquery` INTERSECT SELECT DISTINCT `patient_0001_1`.`patient_id` AS `patient_id` FROM `database_[a-z0-9]{24}__extract_dataset_test`.`patient_0001` AS `patient_0001_1` WHERE `patient_0001_1`.`patient_id` BETWEEN 2 AND 9\)\)\) AS `cohort_query`\);"
        cmd = ["dx", "extract_dataset", cohort_record, "--fields", "patient.patient_id, patient.name,patient.weight, patient.date_of_birth, patient.verified_dtm, test.test_id, trial_visit.visit_id, baseline.baseline_id, hospital.hospital_id , doctor.doctor_id",
               "--sql", "-o", "-"]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        stdout = process.communicate()[0]
        self.assertTrue(re.match(truth_output,stdout.strip()))

    def test_e2e_dataset_fields(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        out_directory = tempfile.mkdtemp()
        cmd = ["dx", "extract_dataset", dataset_record, "--fields", "patient.patient_id, patient.name,patient.weight, patient.date_of_birth, patient.verified_dtm, test.test_id, trial_visit.visit_id, baseline.baseline_id, hospital.hospital_id , doctor.doctor_id",
               "-o", out_directory]
        subprocess.check_call(cmd)
        truth_file = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test.csv"
        self.end_to_end_fields(out_directory=out_directory, rec_name = "extract_dataset_test.csv", truth_file=truth_file)

    def test_e2e_cohortbrowser_fields(self):
        cohort_record = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test"
        out_directory = tempfile.mkdtemp()
        cmd = ["dx", "extract_dataset", cohort_record, "--fields", "patient.patient_id, patient.name,patient.weight, patient.date_of_birth, patient.verified_dtm, test.test_id, trial_visit.visit_id, baseline.baseline_id, hospital.hospital_id , doctor.doctor_id", 
               "-o", out_directory]
        subprocess.check_call(cmd)
        truth_file = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test.csv"
        self.end_to_end_fields(out_directory=out_directory, rec_name = "Combined_Cohort_Test.csv", truth_file=truth_file)
    
    def test_e2e_dataset_fields_file(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        out_directory = tempfile.mkdtemp()
        input_fields = os.path.join(self.test_file_dir , "fields_file.txt")
        cmd = ["dx", "extract_dataset", dataset_record, "--fields-file", input_fields,
               "-o", out_directory]
        subprocess.check_call(cmd)
        truth_file = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test.csv"
        self.end_to_end_fields(out_directory=out_directory, rec_name = "extract_dataset_test.csv", truth_file=truth_file)

    def test_e2e_cohortbrowser_fields_file(self):
        cohort_record = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test"
        out_directory = tempfile.mkdtemp()
        input_fields = os.path.join(self.test_file_dir , "fields_file.txt")
        cmd = ["dx", "extract_dataset", cohort_record, "--fields-file", input_fields, 
               "-o", out_directory]
        subprocess.check_call(cmd)
        truth_file = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test.csv"
        self.end_to_end_fields(out_directory=out_directory, rec_name = "Combined_Cohort_Test.csv", truth_file=truth_file)

    def test_e2e_fields_file_and_fields_negative(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        out_directory = tempfile.mkdtemp()
        expected_error = "dx extract_dataset: error: only one of the arguments, --fields or --fields-file, may be supplied at a given time"
        input_fields = os.path.join(self.test_file_dir , "fields_file.txt")
        cmd = ["dx", "extract_dataset", dataset_record, "--fields", "patient.patient_id", "--fields-file", input_fields,
               "-o", out_directory]
        process = subprocess.Popen(cmd, stderr=subprocess.PIPE, universal_newlines=True)
        stderr = process.communicate()[1]
        self.assertTrue(expected_error in stderr, msg = stderr)

    def test_file_already_exists(self):
        cohort_record = "dx-toolkit_test_data:Extract_Dataset/Combined_Cohort_Test"
        out_directory = tempfile.mkdtemp()
        open(os.path.join(out_directory, "Combined_Cohort_Test.csv"), 'w').close()
        cmd = ["dx", "extract_dataset", cohort_record, "--fields", "patient.patient_id,patient.name", "-o", out_directory]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
        stdout = process.communicate()[0]
        self.assertTrue("Error: path already exists" in stdout.strip())
        shutil.rmtree(out_directory)

    def test_list_entities(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        truth_output = "baseline\tBaseline Test\ndoctor\tDoctors\nhospital\tHospitals\npatient\tPatients\ntest\tTests Performed\ntrial_visit\tVisits"
        cmd = ["dx", "extract_dataset", dataset_record, "--list-entities"]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        stdout = process.communicate()[0]
        self.assertTrue(truth_output in stdout)
    
    def test_list_entities_negative(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        expected_error_contents = ["--list-entities cannot be specified with:", "--sql"]
        cmd = ["dx", "extract_dataset", dataset_record, "--list-entities", "--sql"]
        process = subprocess.Popen(cmd, stderr=subprocess.PIPE, universal_newlines=True)
        stderr = process.communicate()[1]
        for error_content in expected_error_contents:
            self.assertTrue(error_content in stderr)

    def test_list_fields(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        truth_output = "test.result_cat\tResult Class\ntest.results\tRaw Result\ntest.test_date\tTest Date\ntest.test_id\tTest ID\ntest.test_type\tTest\ntest.test_visit_id\ttest_visit_id"
        entities = "test"
        cmd = ["dx", "extract_dataset", dataset_record, "--list-fields", "--entities", entities]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        stdout = process.communicate()[0]
        self.assertTrue(truth_output in stdout)

    def test_list_fields_negative(self):
        dataset_record = "dx-toolkit_test_data:Extract_Dataset/extract_dataset_test"
        expected_error_contents = ["The following entity/entities cannot be found:", "tests"]
        entities = "tests"
        cmd = ["dx", "extract_dataset", dataset_record, "--list-fields","--entities", entities]
        process = subprocess.Popen(cmd, stderr=subprocess.PIPE, universal_newlines=True)
        stderr = process.communicate()[1]
        for error_content in expected_error_contents:
            self.assertTrue(error_content in stderr)

    def end_to_end_ddd(self, out_directory, rec_name):
        truth_files_directory = tempfile.mkdtemp()
        with chdir(truth_files_directory):
            cmd = ["dx", "download", "dx-toolkit_test_data:Extract_Dataset/data_dictionary.csv", 
                                    "dx-toolkit_test_data:Extract_Dataset/codings.csv",
                                    "dx-toolkit_test_data:Extract_Dataset/entity_dictionary.csv"]
            subprocess.check_call(cmd)
            os.chdir("..")
            truth_file_list = os.listdir(truth_files_directory)

            for file in truth_file_list:
                dframe1 = pd.read_csv(os.path.join(truth_files_directory, file)).dropna(axis=1, how='all').sort_index(axis=1)
                fil_nam = rec_name + "." + file
                dframe2 = pd.read_csv(os.path.join(out_directory, fil_nam)).dropna(axis=1, how='all').sort_index(axis=1)
                if file == 'codings.csv':
                    #continue
                    dframe1 = dframe1.sort_values(by=['code','coding_name'], axis=0).reset_index(drop=True)
                    dframe2 = dframe2.sort_values(by=['code','coding_name'], axis=0).reset_index(drop=True)
                elif file in ['entity_dictionary.csv', 'data_dictionary.csv']:
                    dframe1 = dframe1.sort_values(by='entity', axis=0).reset_index(drop=True)
                    dframe2 = dframe2.sort_values(by='entity', axis=0).reset_index(drop=True)
                self.assertTrue(dframe1.equals(dframe2))
            
            shutil.rmtree(out_directory)
            shutil.rmtree(truth_files_directory)

    def end_to_end_fields(self, out_directory, rec_name, truth_file):
        truth_files_directory = tempfile.mkdtemp()
        with chdir(truth_files_directory):
            cmd = ["dx", "download", truth_file]
            subprocess.check_call(cmd)
            os.chdir("..")
            dframe1 = pd.read_csv(os.path.join(truth_files_directory,os.listdir(truth_files_directory)[0]))
            dframe1 = dframe1.sort_values(by=list(dframe1.columns), axis=0).reset_index(drop=True)
            dframe2 = pd.read_csv(os.path.join(out_directory, rec_name))
            dframe2 = dframe2.sort_values(by=list(dframe2.columns), axis=0).reset_index(drop=True)
            self.assertTrue(dframe1.equals(dframe2))

            shutil.rmtree(out_directory)
            shutil.rmtree(truth_files_directory)

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import absolute_import, division, print_function
import contextlib
import codecs
import gzip


@contextlib.contextmanager
def as_handle(path_or_handle, mode=None, is_gzip=False, **kwargs):
    """Open a file path, or pass through an already open file handle.

    Args:
        path_or_handle (str or file-like object):
            The file path to open, or an open file-like object with a 'read'
            method.
        mode (str): File open mode, e.g. 'r' or 'w'
        is_gzip (bool): Whether the file is (or should be) gzip-compressed.
        **kwargs (dict): Passed through to `open`

    Returns: file-like object
    """
    if mode is None:
        mode = 'rb' if is_gzip else 'r'
    if hasattr(path_or_handle, 'read'):
        # File handle is already open
        if is_gzip:
            yield gzip.GzipFile(fileobj=path_or_handle, mode=mode)
        else:
            yield path_or_handle
    else:
        # File path needs to be opened
        if 'encoding' in kwargs:
            opener = codecs.open
        elif is_gzip:
            opener = gzip.open
            # Need to add this for python 3.5
            if "r" in mode: mode = "rt"
        else:
            opener = open
        with opener(path_or_handle, mode=mode, **kwargs) as fp:
            yield fp


#!/usr/bin/env python3
# DX_APP_WIZARD_NAME DX_APP_WIZARD_VERSION
# Generated by dx-app-wizard.
#
# Parallelized execution pattern: Your app will generate multiple jobs
# to perform some computation in parallel, followed by a final
# "postprocess" stage that will perform any additional computations as
# necessary.
#
# See https://documentation.dnanexus.com/developer for documentation and
# tutorials on how to modify this file.
#
# DNAnexus Python Bindings (dxpy) documentation:
#   http://autodoc.dnanexus.com/bindings/python/current/

import os
import dxpy

@dxpy.entry_point("postprocess")
def postprocess(process_outputs):
    # Change the following to process whatever input this stage
    # receives.  You may also want to copy and paste the logic to download
    # and upload files here as well if this stage receives file input
    # and/or makes file output.

    for output in process_outputs:
        pass

    return { "answer": "placeholder value" }

@dxpy.entry_point("process")
def process(input1):
    # Change the following to process whatever input this stage
    # receives.  You may also want to copy and paste the logic to download
    # and upload files here as well if this stage receives file input
    # and/or makes file output.

    print(input1)

    return { "output": "placeholder value" }

@dxpy.entry_point("main")
def main(DX_APP_WIZARD_INPUT_SIGNATURE):
DX_APP_WIZARD_INITIALIZE_INPUTDX_APP_WIZARD_DOWNLOAD_ANY_FILES
    # Split your work into parallel tasks.  As an example, the
    # following generates 10 subjobs running with the same dummy
    # input.

    subjobs = []
    for i in range(10):
        subjob_input = { "input1": True }
        subjobs.append(dxpy.new_dxjob(subjob_input, "process"))

    # The following line creates the job that will perform the
    # "postprocess" step of your app.  We've given it an input field
    # that is a list of job-based object references created from the
    # "process" jobs we just created.  Assuming those jobs have an
    # output field called "output", these values will be passed to the
    # "postprocess" job.  Because these values are not ready until the
    # "process" jobs finish, the "postprocess" job WILL NOT RUN until
    # all job-based object references have been resolved (i.e. the
    # jobs they reference have finished running).
    #
    # If you do not plan to have the "process" jobs create output that
    # the "postprocess" job will require, then you can explicitly list
    # the dependencies to wait for those jobs to finish by setting the
    # "depends_on" field to the list of subjobs to wait for (it
    # accepts either dxpy handlers or string IDs in the list).  We've
    # included this parameter in the line below as well for
    # completeness, though it is unnecessary if you are providing
    # job-based object references in the input that refer to the same
    # set of jobs.

    postprocess_job = dxpy.new_dxjob(fn_input={ "process_outputs": [subjob.get_output_ref("output") for subjob in subjobs] },
                                     fn_name="postprocess",
                                     depends_on=subjobs)
DX_APP_WIZARD_UPLOAD_ANY_FILES
    # If you would like to include any of the output fields from the
    # postprocess_job as the output of your app, you should return it
    # here using a job-based object reference.  If the output field in
    # the postprocess function is called "answer", you can pass that
    # on here as follows:
    #
    # return { "app_output_field": postprocess_job.get_output_ref("answer"), ...}
    #
    # Tip: you can include in your output at this point any open
    # objects (such as files) which will be closed by a job that
    # finishes later.  The system will check to make sure that the
    # output object is closed and will attempt to clone it out as
    # output into the parent container only after all subjobs have
    # finished.

    output = {}
DX_APP_WIZARD_OUTPUT
    return output

dxpy.run()




def transform_to_expression_matrix(list_of_dicts):
    """
    list_of_dicts: list of dictionaries, each of the form
    {
        "feature_id":<feature_id>,
        "sample_id":<sample_id>,
        "expression":<expression>
    }
    Additional key:value pairs are possible, but are not included in the transformed output

    How it works:
    First, create a dictionary where each key is a sample id, and the values are the feature_id:expression pairs
    Then, convert it to a list of dictionaries (the format that output handling is expecting)
    Each item in the list is a dictionary containing one "sample_id":sample_id pair, and
    multiple feature_id:expression pairs

    Returns:
    A list of dictionaries, each of the form:
    {
        "sample_id":<sample_id>,
        <feature_id_1>:<expression_1>
        <feature_id_2>:<expression_2>
        ...
        <feature_id_n>:<expression_n>
    }
    """
    transformed_dict = {}
    # create a dict of the form
    # {
    #   <sample_id>:{
    #       "sample_id":<sample_id>,
    #       <feature_id_1>:<expression_1>,
    #       <feature_id_2>:<expression_2>,
    #        etc.
    #   }
    # 
    # }
    colnames = set()
    for entry in list_of_dicts:
        # Keep track of all seen feature_ids, they will become the columns of our final table
        colnames.add(entry["feature_id"])
        if entry["sample_id"] not in transformed_dict:
            transformed_dict[entry["sample_id"]] = {"sample_id":entry["sample_id"],entry["feature_id"]:entry["expression"]}
        else:
            transformed_dict[entry["sample_id"]][entry["feature_id"]] = entry["expression"]
    colnames = sorted(list(colnames))
    colnames.insert(0,"sample_id")

    # We want the output to be a list of dictionaries, rather than a single dicitonary keyed on sample_id
    dict_list = list(transformed_dict.values())
    for dict_row in dict_list:
        for colname in colnames:
            if colname not in dict_row:
                dict_row[colname] = None

    return (dict_list,colnames)


# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
This submodule contains the callables (and their helpers) that are called by
the org-based commands of the dx command-line client.
'''
from __future__ import print_function, unicode_literals, division, absolute_import

import dxpy
from . import try_call, prompt_for_yn, INTERACTIVE_CLI
from .parsers import process_find_by_property_args, process_phi_param
from ..exceptions import (DXCLIError, err_exit)
from dxpy.utils.printing import (fill, DELIMITER, format_find_results)
import json


def get_user_id(user_id_or_username):
    """Gets the user ID based on the value `user_id_or_username` specified on
    the command-line, being extra lenient and lowercasing the value in all
    cases.
    """
    user_id_or_username = user_id_or_username.lower()
    if not user_id_or_username.startswith("user-"):
        user_id = "user-" + user_id_or_username.lower()
    else:
        user_id = user_id_or_username
    return user_id


def get_org_invite_args(user_id, args):
    """
    Used by:
        - `dx new user`
        - `dx add member`

    PRECONDITION:
        - If /org-x/invite is being called in conjunction with /user/new, then
          `_validate_new_user_input()` has been called on `args`; otherwise,
          the parser must perform all the basic input validation.
    """
    org_invite_args = {"invitee": user_id}
    org_invite_args["level"] = args.level
    if "set_bill_to" in args and args.set_bill_to is True:
        # /org-x/invite is called in conjunction with /user/new.
        org_invite_args["allowBillableActivities"] = True
    else:
        org_invite_args["allowBillableActivities"] = args.allow_billable_activities
    org_invite_args["appAccess"] = args.app_access
    org_invite_args["projectAccess"] = args.project_access
    org_invite_args["suppressEmailNotification"] = args.no_email
    return org_invite_args


def add_membership(args):
    user_id = get_user_id(args.username_or_user_id)

    try:
        dxpy.api.org_find_members(args.org_id, {"id": [user_id]})["results"][0]
    except:
        pass
    else:
        raise DXCLIError("Cannot add a user who is already a member of the org. To update an existing member's permissions, use 'dx update member'")

    dxpy.api.org_invite(args.org_id, get_org_invite_args(user_id, args))

    if args.brief:
        print("org-" + args.org_id)
    else:
        print(fill("Invited {u} to {o}".format(u=user_id, o=args.org_id)))


def _get_org_remove_member_args(args):
    remove_member_args = {
        "user": get_user_id(args.username_or_user_id),
        "revokeProjectPermissions": args.revoke_project_permissions,
        "revokeAppPermissions": args.revoke_app_permissions}
    return remove_member_args


def remove_membership(args):
    user_id = get_user_id(args.username_or_user_id)

    try:
        dxpy.api.org_find_members(args.org_id, {"id": [user_id]})["results"][0]
    except IndexError:
        raise DXCLIError("Cannot remove a user who is not a member of the org")

    confirmed = not args.confirm
    if not confirmed:
        # Request interactive confirmation.
        print(fill("WARNING: About to remove {u} from {o}; project permissions will{rpp} be removed and app permissions will{rap} be removed".format(
            u=user_id, o=args.org_id,
            rpp="" if args.revoke_project_permissions else " not",
            rap="" if args.revoke_app_permissions else " not")))

        if prompt_for_yn("Please confirm"):
            confirmed = True

    if confirmed:
        result = dxpy.api.org_remove_member(args.org_id,
                                            _get_org_remove_member_args(args))
        if args.brief:
            print(result["id"])
        else:
            print(fill("Removed {u} from {o}".format(u=user_id, o=args.org_id)))
            print(fill("Removed {u} from the following projects:".format(u=user_id)))
            if len(result["projects"].keys()) != 0:
                for project_id in result["projects"].keys():
                    print("\t{p}".format(p=project_id))
            else:
                print("\tNone")
            print(fill("Removed {u} from the following apps:".format(u=user_id)))
            if len(result["apps"].keys()) != 0:
                for app_id in result["apps"].keys():
                    print("\t{a}".format(a=app_id))
            else:
                print("\tNone")
    else:
        print(fill("Aborting removal of {u} from {o}".format(u=user_id, o=args.org_id)))


def _get_org_set_member_access_args(args, current_level):
    user_id = get_user_id(args.username_or_user_id)
    org_set_member_access_input = {user_id: {}}

    if args.level is not None:
        org_set_member_access_input[user_id]["level"] = args.level
    else:
        org_set_member_access_input[user_id]["level"] = current_level

    admin_to_member = args.level == "MEMBER" and current_level == "ADMIN"

    if args.allow_billable_activities is not None:
        org_set_member_access_input[user_id]["allowBillableActivities"] = args.allow_billable_activities == "true"
    elif admin_to_member:
        org_set_member_access_input[user_id]["allowBillableActivities"] = False

    if args.app_access is not None:
        org_set_member_access_input[user_id]["appAccess"] = args.app_access == "true"
    elif admin_to_member:
        org_set_member_access_input[user_id]["appAccess"] = True

    if args.project_access is not None:
        org_set_member_access_input[user_id]["projectAccess"] = args.project_access
    elif admin_to_member:
        org_set_member_access_input[user_id]["projectAccess"] = "CONTRIBUTE"

    return org_set_member_access_input


def update_membership(args):
    user_id = get_user_id(args.username_or_user_id)

    try:
        member_access = dxpy.api.org_find_members(args.org_id, {"id": [user_id]})["results"][0]
    except IndexError:
        raise DXCLIError("Cannot update a user who is not a member of the org")

    current_level = member_access["level"]

    result = dxpy.api.org_set_member_access(args.org_id,
                                            _get_org_set_member_access_args(args,
                                                                            current_level))
    if args.brief:
        print(result["id"])
    else:
        print(fill("Updated membership of {u} in {o}".format(u=user_id, o=args.org_id)))


def _get_find_orgs_args(args):
    find_orgs_input = {"level": args.level}

    if args.with_billable_activities is not None:
        find_orgs_input["allowBillableActivities"] = args.with_billable_activities

    if not args.brief:
        find_orgs_input["describe"] = True

    return {"query": find_orgs_input}


def find_orgs(args):
    res_iter = dxpy.find_orgs(_get_find_orgs_args(args)["query"])

    if args.json:
        print(json.dumps(list(res_iter)))
    elif args.brief:
        for res in res_iter:
            print(res["id"])
    else:
        for res in res_iter:
            print("{o}{d1}{n}".format(
                o=res["id"],
                d1=(DELIMITER(args.delimiter) if args.delimiter else " : "),
                n=res["describe"]["name"]
            ))


def org_find_members(args):
    results = try_call(dxpy.org_find_members, org_id=args.org_id, level=args.level, describe=(not args.brief))
    format_find_results(args, results)


def new_org(args):
    if args.name is None and INTERACTIVE_CLI:
        args.name = input("Enter descriptive name for new org: ")

    if args.name is None:
        err_exit("No org name supplied and input is not interactive.")

    org_new_input = {"handle": args.handle, "name": args.name,
                     "policies": {"memberListVisibility": args.member_list_visibility,
                                  "restrictProjectTransfer": args.project_transfer_ability}}

    resp = try_call(dxpy.api.org_new, org_new_input)
    if args.brief:
        print(resp['id'])
    else:
        print('Created new org called "' + args.name + '" (' + resp['id'] + ')')


def _get_org_update_args(args):
    org_update_inputs = {}

    if args.name is not None:
        org_update_inputs["name"] = args.name
    if args.saml_idp is not None:
        org_update_inputs["samlIdP"] = args.saml_idp

    if any(policy not in (None, False) for policy in (args.member_list_visibility, args.project_transfer_ability, args.enable_job_reuse, args.disable_job_reuse)) or args.detailed_job_metrics_collect_default is not None:
        org_update_inputs["policies"] = {}
    if args.member_list_visibility is not None:
        org_update_inputs["policies"]["memberListVisibility"] = args.member_list_visibility
    if args.project_transfer_ability is not None:
        org_update_inputs["policies"]["restrictProjectTransfer"] = args.project_transfer_ability        
    if args.enable_job_reuse == True:
        org_update_inputs["policies"]["jobReuse"] = True
    elif args.disable_job_reuse == True:
        org_update_inputs["policies"]["jobReuse"] = False
    if args.detailed_job_metrics_collect_default is not None:
        org_update_inputs["policies"]["detailedJobMetricsCollectDefault"] = args.detailed_job_metrics_collect_default == 'true'
    if args.job_logs_forwarding_json is not None:
        try: 
            org_update_inputs["jobLogsForwarding"] = json.loads(args.job_logs_forwarding_json)
        except:
            err_exit("Invalid JSON input for --job-logs-forwarding-json")

    return org_update_inputs


def update_org(args):
    org_update_inputs = _get_org_update_args(args)
    res = try_call(dxpy.api.org_update, args.org_id, org_update_inputs)
    if args.brief:
        print(res["id"])
    else:
        print(fill("Updated {o}".format(o=res["id"])))


def org_find_projects(args):
    try_call(process_find_by_property_args, args)
    try_call(process_phi_param, args)
    try:
        results = dxpy.org_find_projects(org_id=args.org_id, name=args.name, name_mode='glob',
                                         ids=args.ids, properties=args.properties, tags=args.tag,
                                         describe=(not args.brief),
                                         public=args.public,
                                         created_after=args.created_after,
                                         created_before=args.created_before,
                                         region=args.region,
                                         containsPHI=args.containsPHI)
    except:
        err_exit()
    format_find_results(args, results)


def org_find_apps(args):
    try:
        results = dxpy.org_find_apps(org_id=args.org_id,
                                     name=args.name,
                                     name_mode='glob',
                                     describe=(not args.brief),
                                     created_after=args.created_after,
                                     created_before=args.created_before)
    except:
        err_exit()
    format_find_results(args, results)



# Copyright (C) 2014-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
Adapted from http://norvig.com/spell-correct.html.
"""

from __future__ import print_function, unicode_literals, division, absolute_import

import re, collections

def _train(features):
    model = collections.defaultdict(lambda: 1)
    for f in features:
        model[f] += 1
    return model

_alphabet = 'abcdefghijklmnopqrstuvwxyz'

def _edits1(word):
   splits     = [(word[:i], word[i:]) for i in range(len(word) + 1)]
   deletes    = [a + b[1:] for a, b in splits if b]
   transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b)>1]
   replaces   = [a + c + b[1:] for a, b in splits for c in _alphabet if b]
   inserts    = [a + c + b     for a, b in splits for c in _alphabet]
   return set(deletes + transposes + replaces + inserts)

def _known_edits2(word, NWORDS):
    return set(e2 for e1 in _edits1(word) for e2 in _edits1(e1) if e2 in NWORDS)

def _known(words, NWORDS):
    return set(w for w in words if w in NWORDS)

def correct(word, known_words):
    """
    :param word: Word to correct
    :type word: string
    :param known_words: List of known words
    :type known_words: iterable of strings

    Given **word**, suggests a correction from **known_words**. If no reasonably close correction is found, returns
    **word**.
    """
    NWORDS = _train(known_words)
    candidates = _known([word], NWORDS) or _known(_edits1(word), NWORDS) or _known_edits2(word, NWORDS) or [word]
    return max(candidates, key=NWORDS.get)
